
#line 3 "lex.yy.c"

#define  YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 5
#define YY_FLEX_SUBMINOR_VERSION 37
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

/* end standard C headers. */

/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types. 
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <inttypes.h>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t; 
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN               (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN              (-32767-1)
#endif
#ifndef INT32_MIN
#define INT32_MIN              (-2147483647-1)
#endif
#ifndef INT8_MAX
#define INT8_MAX               (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX              (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX              (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX              (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX             (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX             (4294967295U)
#endif

#endif /* ! C99 */

#endif /* ! FLEXINT_H */

#ifdef __cplusplus

/* The "const" storage-class-modifier is valid. */
#define YY_USE_CONST

#else	/* ! __cplusplus */

/* C99 requires __STDC__ to be defined as 1. */
#if defined (__STDC__)

#define YY_USE_CONST

#endif	/* defined (__STDC__) */
#endif	/* ! __cplusplus */

#ifdef YY_USE_CONST
#define yyconst const
#else
#define yyconst
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an unsigned
 * integer for use as an array index.  If the signed char is negative,
 * we want to instead treat it as an 8-bit unsigned char, hence the
 * double cast.
 */
#define YY_SC_TO_UI(c) ((unsigned int) (unsigned char) c)

/* An opaque pointer. */
#ifndef YY_TYPEDEF_YY_SCANNER_T
#define YY_TYPEDEF_YY_SCANNER_T
typedef void* yyscan_t;
#endif

/* For convenience, these vars (plus the bison vars far below)
   are macros in the reentrant scanner. */
#define yyin yyg->yyin_r
#define yyout yyg->yyout_r
#define yyextra yyg->yyextra_r
#define yyleng yyg->yyleng_r
#define yytext yyg->yytext_r
#define yylineno (YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno)
#define yycolumn (YY_CURRENT_BUFFER_LVALUE->yy_bs_column)
#define yy_flex_debug yyg->yy_flex_debug_r

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN yyg->yy_start = 1 + 2 *

/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START ((yyg->yy_start - 1) / 2)
#define YYSTATE YY_START

/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)

/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yyrestart(yyin ,yyscanner )

#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#define YY_BUF_SIZE 16384
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

    #define YY_LESS_LINENO(n)
    
/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		*yy_cp = yyg->yy_hold_char; \
		YY_RESTORE_YY_MORE_OFFSET \
		yyg->yy_c_buf_p = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up yytext again */ \
		} \
	while ( 0 )

#define unput(c) yyunput( c, yyg->yytext_ptr , yyscanner )

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
	{
	FILE *yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	yy_size_t yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	yy_size_t yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

    int yy_bs_lineno; /**< The line count. */
    int yy_bs_column; /**< The column count. */
    
	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yyrestart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2

	};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER ( yyg->yy_buffer_stack \
                          ? yyg->yy_buffer_stack[yyg->yy_buffer_stack_top] \
                          : NULL)

/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE yyg->yy_buffer_stack[yyg->yy_buffer_stack_top]

void yyrestart (FILE *input_file ,yyscan_t yyscanner );
void yy_switch_to_buffer (YY_BUFFER_STATE new_buffer ,yyscan_t yyscanner );
YY_BUFFER_STATE yy_create_buffer (FILE *file,int size ,yyscan_t yyscanner );
void yy_delete_buffer (YY_BUFFER_STATE b ,yyscan_t yyscanner );
void yy_flush_buffer (YY_BUFFER_STATE b ,yyscan_t yyscanner );
void yypush_buffer_state (YY_BUFFER_STATE new_buffer ,yyscan_t yyscanner );
void yypop_buffer_state (yyscan_t yyscanner );

static void yyensure_buffer_stack (yyscan_t yyscanner );
static void yy_load_buffer_state (yyscan_t yyscanner );
static void yy_init_buffer (YY_BUFFER_STATE b,FILE *file ,yyscan_t yyscanner );

#define YY_FLUSH_BUFFER yy_flush_buffer(YY_CURRENT_BUFFER ,yyscanner)

YY_BUFFER_STATE yy_scan_buffer (char *base,yy_size_t size ,yyscan_t yyscanner );
YY_BUFFER_STATE yy_scan_string (yyconst char *yy_str ,yyscan_t yyscanner );
YY_BUFFER_STATE yy_scan_bytes (yyconst char *bytes,yy_size_t len ,yyscan_t yyscanner );

void *yyalloc (yy_size_t ,yyscan_t yyscanner );
void *yyrealloc (void *,yy_size_t ,yyscan_t yyscanner );
void yyfree (void * ,yyscan_t yyscanner );

#define yy_new_buffer yy_create_buffer

#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){ \
        yyensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer(yyin,YY_BUF_SIZE ,yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive; \
	}

#define yy_set_bol(at_bol) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){\
        yyensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer(yyin,YY_BUF_SIZE ,yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol; \
	}

#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* Begin user sect3 */

#define yywrap(yyscanner) 1
#define YY_SKIP_YYWRAP

typedef unsigned char YY_CHAR;

typedef int yy_state_type;

#define yytext_ptr yytext_r

static yy_state_type yy_get_previous_state (yyscan_t yyscanner );
static yy_state_type yy_try_NUL_trans (yy_state_type current_state  ,yyscan_t yyscanner);
static int yy_get_next_buffer (yyscan_t yyscanner );
static void yy_fatal_error (yyconst char msg[] ,yyscan_t yyscanner );

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	yyg->yytext_ptr = yy_bp; \
	yyleng = (size_t) (yy_cp - yy_bp); \
	yyg->yy_hold_char = *yy_cp; \
	*yy_cp = '\0'; \
	yyg->yy_c_buf_p = yy_cp;

#define YY_NUM_RULES 72
#define YY_END_OF_BUFFER 73
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
	{
	flex_int32_t yy_verify;
	flex_int32_t yy_nxt;
	};
static yyconst flex_int16_t yy_accept[241] =
    {   0,
        0,    0,   73,    7,    7,    7,    1,    2,    5,    6,
       72,   72,   72,   72,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   18,   20,   19,    0,    0,    0,
        4,    0,    0,    0,    0,    0,    0,   17,    0,    0,
        0,    0,    0,    0,    3,   13,    0,   21,   22,    7,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,   58,   59,    0,    0,   62,    0,    0,
        0,    0,    0,    0,    0,   60,    0,    0,    0,   12,
        0,    0,    0,    0,    0,   10,    8,    9,   11,   47,

       48,   50,   51,   67,   49,   54,   55,   56,   57,   61,
       69,   63,   64,   65,   66,   70,   68,   53,   52,   23,
       24,   36,   26,   27,   43,   25,   30,   31,   32,   33,
       34,   35,   37,   38,   45,   39,   40,   41,   42,   46,
       44,   29,   28,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,   16,    0,    0,    0,    0,    0,    0,    0,    0,
       14,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,

        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,   23,    0,    0,   15,
        0,   13,    0,    0,    0,    0,    0,    0,    0,    0
    } ;

static yyconst flex_int32_t yy_ec[256] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        2,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    3,
        4,    1,    1,    1,    1,    1,    1,    1,    5,    6,
        1,    1,    1,    7,    1,    8,    2,    2,    2,    2,
        2,    2,    2,    2,    9,   10,   11,   12,   13,   14,
       15,   16,   17,    2,   18,   19,   20,   21,   22,   23,
       24,   25,   26,   27,   28,    2,   29,   30,   31,   32,
        2,   33,    2,   34,   35,   36,   37,   38,   39,   40,

       41,   42,   43,   44,   45,   36,   46,   47,   48,   49,
       50,   51,   52,   53,   54,   55,   56,   36,   57,   58,
       59,   60,   61,   62,   63,   64,   64,   65,   65,   65,
       65,   65,   65,   65,   65,   65,   65,   65,   65,   65,
       65,   65,   65,   65,   65,   65,   65,   65,   65,   65,
       65,   65,   65,   65,   65,   65,   65,   65,   65,   66,
       66,   66,   66,   66,   66,   66,   66,   66,   66,   66,
       66,   66,   66,   66,   66,   66,   66,   66,   66,   66,
       66,   66,   66,   66,   66,   66,   66,   66,   66,   66,
       66,   67,   67,   68,   68,   68,   68,   68,   68,   68,

       68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
       68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
       68,   68,   68,   69,   70,   70,   70,   70,   70,   70,
       70,   70,   70,   70,   70,   70,   70,   70,   70,   67,
       67,   67,   67,   67,   67,   67,   67,   67,   67,   67,
       67,   67,   67,   67,   67
    } ;

static yyconst flex_int32_t yy_meta[71] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1
    } ;

static yyconst flex_int16_t yy_base[241] =
    {   0,
        0,    0,  357,  358,  358,   68,  358,  358,  358,  358,
      358,    7,  290,    9,  349,  307,  312,  308,  310,   79,
      313,  299,  311,  310,  290,  289,  296,   88,  299,   80,
       55,  291,  296,  299,  358,  358,  358,  292,   96,   60,
      297,   84,  102,  148,   94,  300,  299,  279,   90,   36,
      165,   97,  289,  176,  282,  287,  290,  358,  358,  358,
       87,  116,  324,  278,  273,  282,  279,  271,  287,  275,
      267,  270,  272,  358,  358,  109,  274,  358,  273,  267,
      273,  266,  258,  272,  258,  358,  256,  259,  254,  358,
      261,  262,  259,  251,  267,  358,  358,  358,  358,  358,

      358,  358,  358,  358,  358,  358,  358,  358,  358,  358,
      358,  358,  358,  358,  358,  358,  358,  358,  358,  255,
      358,  358,  358,  358,  358,  358,  358,  358,  358,  358,
      358,  358,  358,  358,  358,  358,  358,  358,  358,  358,
      358,  358,  358,  247,  246,  249,  251,  260,  247,  170,
      251,  250,  244,  250,  236,  250,  242,  235,  233,  176,
      243,  249,  230,  239,  235,  245,  230,  242,  236,  239,
      229,  358,  221,  230,  237,  229,  235,  216,  225,  221,
      358,  231,  216,  228,  218,  209,  220,  223,  213,  205,
      218,  213,  220,  209,  251,  218,  217,  206,  215,  214,

      210,  212,  195,  210,  209,  198,  207,  206,  195,  204,
      203,  199,  201,  200,  183,  198,  197,  179,  185,  227,
      226,  225,  178,  190,  176,  175,  358,  174,  186,  358,
      172,  358,  168,  165,  163,   99,   71,   63,   33,  358
    } ;

static yyconst flex_int16_t yy_def[241] =
    {   0,
      240,    1,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,

      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,

      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,    0
    } ;

static yyconst flex_int16_t yy_nxt[429] =
    {   0,
        4,    5,    4,    4,    4,    4,    4,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    6,    7,    8,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        9,    4,   10,    4,   11,   11,   11,   12,   13,   14,
       15,   60,   60,   62,   62,  149,   16,   17,   18,   19,
       20,  139,   21,  150,   22,   23,   24,   25,   26,   27,
       28,   83,   29,   30,   31,   32,   91,   33,   84,   34,

       35,   36,   37,   92,   38,   39,   40,   41,   42,   43,
       44,  133,   45,   46,   47,   48,   49,   50,   51,  124,
       52,   53,   54,   55,   81,   56,  148,   57,   58,   68,
       59,   77,   78,   69,   94,   82,   89,   90,   95,   90,
      153,   79,  144,  145,   98,  132,   96,  115,   96,  169,
       97,   60,   60,  170,   98,   99,  100,  101,   86,  102,
      103,  104,  105,  106,  107,  108,  109,   74,   75,  110,
       78,  111,  112,  113,  114,  115,  116,  117,  118,  119,
       60,   60,  194,  195,  120,  121,  122,  123,  124,  125,
      126,  127,  128,  129,  130,  131,  132,  133,  134,  135,

      136,  137,  138,  139,  140,  141,  142,  143,  151,  134,
      187,  110,  155,  103,  188,  220,  221,  239,  152,  156,
      157,  238,  130,  237,  236,  235,  109,  234,  172,  181,
      230,  233,  232,  127,  137,  231,  140,  230,  229,  129,
      126,  228,  123,  227,  226,  106,  113,  225,  116,  224,
      108,  105,  223,  102,  100,  222,  143,  219,  218,  217,
      216,  215,  214,  172,  213,  212,  211,  128,  210,  209,
      208,  121,  207,  119,  206,  205,  204,  203,  202,  201,
      200,  107,  199,  198,  197,  101,  196,  193,  192,  191,
      190,  138,  189,  136,  135,  125,  186,  185,  184,  183,

      182,  181,  180,  142,  179,  178,  141,   99,  177,  176,
      175,  174,  173,  114,  172,  171,  112,  111,  104,  168,
      167,  166,  165,  118,  164,  163,  117,  162,  161,  160,
      159,  122,  158,  154,  131,  147,  146,   93,   88,   87,
       86,   85,   80,   76,   75,   74,   73,   72,   71,   70,
       67,   66,   65,   64,   63,   61,  240,    3,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,

      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240
    } ;

static yyconst flex_int16_t yy_chk[429] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        6,   12,   12,   14,   14,   50,    6,    6,    6,    6,
        6,  239,    6,   50,    6,    6,    6,    6,    6,    6,
        6,   31,    6,    6,    6,    6,   40,    6,   31,    6,

        6,    6,    6,   40,    6,    6,    6,    6,    6,    6,
        6,  238,    6,    6,    6,    6,    6,    6,    6,  237,
        6,    6,    6,    6,   30,    6,   49,    6,    6,   20,
        6,   28,   28,   20,   42,   30,   39,   39,   42,   43,
       52,   28,   45,   45,   52,   49,   43,  236,   45,   76,
       43,   61,   61,   76,   43,   43,   44,   44,   44,   44,
       44,   44,   44,   44,   44,   44,   44,   44,   44,   44,
       44,   44,   44,   44,   44,   44,   44,   44,   44,   44,
       62,   62,  160,  160,   44,   44,   44,   44,   44,   44,
       44,   44,   44,   44,   44,   44,   44,   44,   44,   44,

       44,   44,   44,   44,   44,   44,   44,   44,   51,   51,
      150,  235,   54,  234,  150,  194,  194,  233,   51,   54,
       54,  231,  229,  228,  226,  225,  224,  223,  222,  221,
      220,  219,  218,  217,  216,  215,  214,  213,  212,  211,
      210,  209,  208,  207,  206,  205,  204,  203,  202,  201,
      200,  199,  198,  197,  196,  195,  193,  192,  191,  190,
      189,  188,  187,  186,  185,  184,  183,  182,  180,  179,
      178,  177,  176,  175,  174,  173,  171,  170,  169,  168,
      167,  166,  165,  164,  163,  162,  161,  159,  158,  157,
      156,  155,  154,  153,  152,  151,  149,  148,  147,  146,

      145,  144,  120,   95,   94,   93,   92,   91,   89,   88,
       87,   85,   84,   83,   82,   81,   80,   79,   77,   73,
       72,   71,   70,   69,   68,   67,   66,   65,   64,   63,
       57,   56,   55,   53,   48,   47,   46,   41,   38,   34,
       33,   32,   29,   27,   26,   25,   24,   23,   22,   21,
       19,   18,   17,   16,   15,   13,    3,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,

      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
      240,  240,  240,  240,  240,  240,  240,  240
    } ;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
#line 1 "giza-scanner.l"
/* A re-entrant scanner to recursively print text */
/* Specify that it is re-entrant and doesn't use ywrap */
/* Include the header and define tokens */
#line 9 "giza-scanner.l"
/* giza - a scientific plotting library built on cairo
 *
 * Copyright (c) 2010      James Wetter and Daniel Price
 * Copyright (c) 2010-2014 Daniel Price
 *
 * This library is free software; and you are welcome to redistribute
 * it under the terms of the GNU General Public License
 * (GPL, see LICENSE file for details) and the provision that
 * this notice remains intact. If you modify this file, please
 * note section 2a) of the GPLv2 states that:
 *
 *  a) You must cause the modified files to carry prominent notices
 *     stating that you changed the files and the date of any change.
 *
 * This software is distributed "AS IS", with ABSOLUTELY NO WARRANTY.
 * See the GPL for specific language governing rights and limitations.
 *
 * The Original code is the giza plotting library.
 *
 * Contributor(s):
 *      James Wetter <wetter.j@gmail.com>
 *      Daniel Price <daniel.price@monash.edu> (main contact)
 */
#include "giza-private.h"
#include "giza-transforms-private.h"
#include "giza-drivers-private.h"
#include "giza-text-private.h"
#include "giza.h"
#include <string.h>
#include <math.h>

#define GIZA_TOKEN_ERROR  0
#define GIZA_TOKEN_END    1
#define GIZA_TOKEN_SUPER  2
#define GIZA_TOKEN_SUB    3
#define GIZA_TOKEN_OB     4
#define GIZA_TOKEN_CB     5
#define GIZA_TOKEN_TIMES  6
#define GIZA_TOKEN_OTHER  7
#define GIZA_TOKEN_RAISE  8
#define GIZA_TOKEN_LOWER  9
#define GIZA_TOKEN_INTEGRAL 10
#define GIZA_TOKEN_NABLA 11

#define GIZA_TOKEN_FONT_NORMAL 12
#define GIZA_TOKEN_FONT_ROMAN 13
#define GIZA_TOKEN_FONT_ITALIC 14
#define GIZA_TOKEN_FONT_SCRIPT 15
#define GIZA_TOKEN_FONT_BOLD 16

#define GIZA_TOKEN_ALPHA   21
#define GIZA_TOKEN_BETA    22
#define GIZA_TOKEN_GAMMA   23
#define GIZA_TOKEN_DELTA   24
#define GIZA_TOKEN_EPSILON 25
#define GIZA_TOKEN_ZETA    26
#define GIZA_TOKEN_ETA     27
#define GIZA_TOKEN_THETA   28
#define GIZA_TOKEN_IOTA    29
#define GIZA_TOKEN_KAPPA   30
#define GIZA_TOKEN_LAMBDA  31
#define GIZA_TOKEN_MU      32
#define GIZA_TOKEN_NU      33
#define GIZA_TOKEN_XI      34
#define GIZA_TOKEN_OMICRON 35
#define GIZA_TOKEN_PI      36
#define GIZA_TOKEN_RHO     37
#define GIZA_TOKEN_SIGMA   38
#define GIZA_TOKEN_TAU     39
#define GIZA_TOKEN_UPSILON 40
#define GIZA_TOKEN_PHI     41
#define GIZA_TOKEN_CHI     42
#define GIZA_TOKEN_PSI     43
#define GIZA_TOKEN_OMEGA   44

#define GIZA_TOKEN_BIGALPHA   45
#define GIZA_TOKEN_BIGBETA    46
#define GIZA_TOKEN_BIGGAMMA   47
#define GIZA_TOKEN_BIGDELTA   48
#define GIZA_TOKEN_BIGEPSILON 49
#define GIZA_TOKEN_BIGZETA    50
#define GIZA_TOKEN_BIGETA     51
#define GIZA_TOKEN_BIGTHETA   52
#define GIZA_TOKEN_BIGIOTA    53
#define GIZA_TOKEN_BIGKAPPA   54
#define GIZA_TOKEN_BIGLAMBDA  55
#define GIZA_TOKEN_BIGMU      56
#define GIZA_TOKEN_BIGNU      57
#define GIZA_TOKEN_BIGXI      58
#define GIZA_TOKEN_BIGOMICRON 59
#define GIZA_TOKEN_BIGPI      60
#define GIZA_TOKEN_BIGRHO     61
#define GIZA_TOKEN_BIGSIGMA   62
#define GIZA_TOKEN_BIGTAU     63
#define GIZA_TOKEN_BIGUPSILON 64
#define GIZA_TOKEN_BIGPHI     65
#define GIZA_TOKEN_BIGCHI     66
#define GIZA_TOKEN_BIGPSI     67
#define GIZA_TOKEN_BIGOMEGA   68

#define GIZA_TOKEN_ODOT 69
#define GIZA_TOKEN_MARKER 70
#define GIZA_TOKEN_BACKSLASH 71
#define GIZA_TOKEN_UNDERSCORE 72
#define GIZA_TOKEN_HAT 73
#define GIZA_TOKEN_CURLYOPEN 74
#define GIZA_TOKEN_CURLYCLOSE 75

#define GIZA_SCALE_SUPERS 0.65
#define GIZA_RAISE_SUPERS 0.4
#define GIZA_SCALE_SUBS   0.65
#define GIZA_LOWER_SUBS   0.2

void _giza_add_token_to_string (char *string, yyscan_t scanner, int token);
int _giza_get_chunk (const char *text);

#line 725 "lex.yy.c"

#define INITIAL 0

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
#include <unistd.h>
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

/* Holds the entire state of the reentrant scanner. */
struct yyguts_t
    {

    /* User-defined. Not touched by flex. */
    YY_EXTRA_TYPE yyextra_r;

    /* The rest are the same as the globals declared in the non-reentrant scanner. */
    FILE *yyin_r, *yyout_r;
    size_t yy_buffer_stack_top; /**< index of top of stack. */
    size_t yy_buffer_stack_max; /**< capacity of stack. */
    YY_BUFFER_STATE * yy_buffer_stack; /**< Stack as an array. */
    char yy_hold_char;
    yy_size_t yy_n_chars;
    yy_size_t yyleng_r;
    char *yy_c_buf_p;
    int yy_init;
    int yy_start;
    int yy_did_buffer_switch_on_eof;
    int yy_start_stack_ptr;
    int yy_start_stack_depth;
    int *yy_start_stack;
    yy_state_type yy_last_accepting_state;
    char* yy_last_accepting_cpos;

    int yylineno_r;
    int yy_flex_debug_r;

    char *yytext_r;
    int yy_more_flag;
    int yy_more_len;

    }; /* end struct yyguts_t */

static int yy_init_globals (yyscan_t yyscanner );

int yylex_init (yyscan_t* scanner);

int yylex_init_extra (YY_EXTRA_TYPE user_defined,yyscan_t* scanner);

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int yylex_destroy (yyscan_t yyscanner );

int yyget_debug (yyscan_t yyscanner );

void yyset_debug (int debug_flag ,yyscan_t yyscanner );

YY_EXTRA_TYPE yyget_extra (yyscan_t yyscanner );

void yyset_extra (YY_EXTRA_TYPE user_defined ,yyscan_t yyscanner );

FILE *yyget_in (yyscan_t yyscanner );

void yyset_in  (FILE * in_str ,yyscan_t yyscanner );

FILE *yyget_out (yyscan_t yyscanner );

void yyset_out  (FILE * out_str ,yyscan_t yyscanner );

yy_size_t yyget_leng (yyscan_t yyscanner );

char *yyget_text (yyscan_t yyscanner );

int yyget_lineno (yyscan_t yyscanner );

void yyset_lineno (int line_number ,yyscan_t yyscanner );

int yyget_column  (yyscan_t yyscanner );

void yyset_column (int column_no ,yyscan_t yyscanner );

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yywrap (yyscan_t yyscanner );
#else
extern int yywrap (yyscan_t yyscanner );
#endif
#endif

    static void yyunput (int c,char *buf_ptr  ,yyscan_t yyscanner);
    
#ifndef yytext_ptr
static void yy_flex_strncpy (char *,yyconst char *,int ,yyscan_t yyscanner);
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * ,yyscan_t yyscanner);
#endif

#ifndef YY_NO_INPUT

#ifdef __cplusplus
static int yyinput (yyscan_t yyscanner );
#else
static int input (yyscan_t yyscanner );
#endif

#endif

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#define YY_READ_BUF_SIZE 8192
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO do { if (fwrite( yytext, yyleng, 1, yyout )) {} } while (0)
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if ( YY_CURRENT_BUFFER_LVALUE->yy_is_interactive ) \
		{ \
		int c = '*'; \
		size_t n; \
		for ( n = 0; n < max_size && \
			     (c = getc( yyin )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( yyin ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else \
		{ \
		errno=0; \
		while ( (result = fread(buf, 1, max_size, yyin))==0 && ferror(yyin)) \
			{ \
			if( errno != EINTR) \
				{ \
				YY_FATAL_ERROR( "input in flex scanner failed" ); \
				break; \
				} \
			errno=0; \
			clearerr(yyin); \
			} \
		}\
\

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error( msg , yyscanner)
#endif

/* end tables serialization structures and prototypes */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1

extern int yylex (yyscan_t yyscanner);

#define YY_DECL int yylex (yyscan_t yyscanner)
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK break;
#endif

#define YY_RULE_SETUP \
	YY_USER_ACTION

/** The main scanner function which does all the work.
 */
YY_DECL
{
	register yy_state_type yy_current_state;
	register char *yy_cp, *yy_bp;
	register int yy_act;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

#line 205 "giza-scanner.l"


#line 953 "lex.yy.c"

	if ( !yyg->yy_init )
		{
		yyg->yy_init = 1;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! yyg->yy_start )
			yyg->yy_start = 1;	/* first start state */

		if ( ! yyin )
			yyin = stdin;

		if ( ! yyout )
			yyout = stdout;

		if ( ! YY_CURRENT_BUFFER ) {
			yyensure_buffer_stack (yyscanner);
			YY_CURRENT_BUFFER_LVALUE =
				yy_create_buffer(yyin,YY_BUF_SIZE ,yyscanner);
		}

		yy_load_buffer_state(yyscanner );
		}

	while ( 1 )		/* loops until end-of-file is reached */
		{
		yy_cp = yyg->yy_c_buf_p;

		/* Support of yytext. */
		*yy_cp = yyg->yy_hold_char;

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

		yy_current_state = yyg->yy_start;
yy_match:
		do
			{
			register YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)];
			if ( yy_accept[yy_current_state] )
				{
				yyg->yy_last_accepting_state = yy_current_state;
				yyg->yy_last_accepting_cpos = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 241 )
					yy_c = yy_meta[(unsigned int) yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
			++yy_cp;
			}
		while ( yy_base[yy_current_state] != 358 );

yy_find_action:
		yy_act = yy_accept[yy_current_state];
		if ( yy_act == 0 )
			{ /* have to back up */
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			yy_act = yy_accept[yy_current_state];
			}

		YY_DO_BEFORE_ACTION;

do_action:	/* This label is used only to access EOF actions. */

		switch ( yy_act )
	{ /* beginning of action switch */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = yyg->yy_hold_char;
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			goto yy_find_action;

case 1:
YY_RULE_SETUP
#line 207 "giza-scanner.l"
{ return GIZA_TOKEN_SUPER; }
	YY_BREAK
case 2:
YY_RULE_SETUP
#line 209 "giza-scanner.l"
{ return GIZA_TOKEN_SUB; }
	YY_BREAK
case 3:
YY_RULE_SETUP
#line 211 "giza-scanner.l"
{ return GIZA_TOKEN_RAISE; }
	YY_BREAK
case 4:
YY_RULE_SETUP
#line 213 "giza-scanner.l"
{ return GIZA_TOKEN_LOWER; }
	YY_BREAK
case 5:
YY_RULE_SETUP
#line 215 "giza-scanner.l"
{ return GIZA_TOKEN_OB; }
	YY_BREAK
case 6:
YY_RULE_SETUP
#line 217 "giza-scanner.l"
{ return GIZA_TOKEN_CB; }
	YY_BREAK
case 7:
/* rule 7 can match eol */
YY_RULE_SETUP
#line 219 "giza-scanner.l"
{ return GIZA_TOKEN_OTHER; }
	YY_BREAK
case 8:
YY_RULE_SETUP
#line 221 "giza-scanner.l"
{ return GIZA_TOKEN_FONT_NORMAL; }
	YY_BREAK
case 9:
YY_RULE_SETUP
#line 222 "giza-scanner.l"
{ return GIZA_TOKEN_FONT_ROMAN; }
	YY_BREAK
case 10:
YY_RULE_SETUP
#line 223 "giza-scanner.l"
{ return GIZA_TOKEN_FONT_ITALIC; }
	YY_BREAK
case 11:
YY_RULE_SETUP
#line 224 "giza-scanner.l"
{ return GIZA_TOKEN_FONT_SCRIPT; }
	YY_BREAK
case 12:
YY_RULE_SETUP
#line 225 "giza-scanner.l"
{ return GIZA_TOKEN_FONT_BOLD; }
	YY_BREAK
case 13:
YY_RULE_SETUP
#line 227 "giza-scanner.l"
{ return GIZA_TOKEN_TIMES; }
	YY_BREAK
case 14:
YY_RULE_SETUP
#line 228 "giza-scanner.l"
{ return GIZA_TOKEN_INTEGRAL; }
	YY_BREAK
case 15:
YY_RULE_SETUP
#line 229 "giza-scanner.l"
{ return GIZA_TOKEN_NABLA; }
	YY_BREAK
case 16:
YY_RULE_SETUP
#line 230 "giza-scanner.l"
{ return GIZA_TOKEN_ODOT; }
	YY_BREAK
case 17:
YY_RULE_SETUP
#line 231 "giza-scanner.l"
{ return GIZA_TOKEN_MARKER; }
	YY_BREAK
case 18:
YY_RULE_SETUP
#line 232 "giza-scanner.l"
{ return GIZA_TOKEN_BACKSLASH; }
	YY_BREAK
case 19:
YY_RULE_SETUP
#line 233 "giza-scanner.l"
{ return GIZA_TOKEN_UNDERSCORE; }
	YY_BREAK
case 20:
YY_RULE_SETUP
#line 234 "giza-scanner.l"
{ return GIZA_TOKEN_HAT; }
	YY_BREAK
case 21:
YY_RULE_SETUP
#line 235 "giza-scanner.l"
{ return GIZA_TOKEN_CURLYOPEN; }
	YY_BREAK
case 22:
YY_RULE_SETUP
#line 236 "giza-scanner.l"
{ return GIZA_TOKEN_CURLYCLOSE; }
	YY_BREAK
case 23:
YY_RULE_SETUP
#line 238 "giza-scanner.l"
{ return GIZA_TOKEN_ALPHA; }
	YY_BREAK
case 24:
YY_RULE_SETUP
#line 239 "giza-scanner.l"
{ return GIZA_TOKEN_BETA; }
	YY_BREAK
case 25:
YY_RULE_SETUP
#line 240 "giza-scanner.l"
{ return GIZA_TOKEN_GAMMA; }
	YY_BREAK
case 26:
YY_RULE_SETUP
#line 241 "giza-scanner.l"
{ return GIZA_TOKEN_DELTA; }
	YY_BREAK
case 27:
YY_RULE_SETUP
#line 242 "giza-scanner.l"
{ return GIZA_TOKEN_EPSILON; }
	YY_BREAK
case 28:
YY_RULE_SETUP
#line 243 "giza-scanner.l"
{ return GIZA_TOKEN_ZETA; }
	YY_BREAK
case 29:
YY_RULE_SETUP
#line 244 "giza-scanner.l"
{ return GIZA_TOKEN_ETA; }
	YY_BREAK
case 30:
YY_RULE_SETUP
#line 245 "giza-scanner.l"
{ return GIZA_TOKEN_THETA; }
	YY_BREAK
case 31:
YY_RULE_SETUP
#line 246 "giza-scanner.l"
{ return GIZA_TOKEN_IOTA; }
	YY_BREAK
case 32:
YY_RULE_SETUP
#line 247 "giza-scanner.l"
{ return GIZA_TOKEN_KAPPA; }
	YY_BREAK
case 33:
YY_RULE_SETUP
#line 248 "giza-scanner.l"
{ return GIZA_TOKEN_LAMBDA; }
	YY_BREAK
case 34:
YY_RULE_SETUP
#line 249 "giza-scanner.l"
{ return GIZA_TOKEN_MU; }
	YY_BREAK
case 35:
YY_RULE_SETUP
#line 250 "giza-scanner.l"
{ return GIZA_TOKEN_NU; }
	YY_BREAK
case 36:
YY_RULE_SETUP
#line 251 "giza-scanner.l"
{ return GIZA_TOKEN_XI; }
	YY_BREAK
case 37:
YY_RULE_SETUP
#line 252 "giza-scanner.l"
{ return GIZA_TOKEN_OMICRON; }
	YY_BREAK
case 38:
YY_RULE_SETUP
#line 253 "giza-scanner.l"
{ return GIZA_TOKEN_PI; }
	YY_BREAK
case 39:
YY_RULE_SETUP
#line 254 "giza-scanner.l"
{ return GIZA_TOKEN_RHO; }
	YY_BREAK
case 40:
YY_RULE_SETUP
#line 255 "giza-scanner.l"
{ return GIZA_TOKEN_SIGMA; }
	YY_BREAK
case 41:
YY_RULE_SETUP
#line 256 "giza-scanner.l"
{ return GIZA_TOKEN_TAU; }
	YY_BREAK
case 42:
YY_RULE_SETUP
#line 257 "giza-scanner.l"
{ return GIZA_TOKEN_UPSILON; }
	YY_BREAK
case 43:
YY_RULE_SETUP
#line 258 "giza-scanner.l"
{ return GIZA_TOKEN_PHI; }
	YY_BREAK
case 44:
YY_RULE_SETUP
#line 259 "giza-scanner.l"
{ return GIZA_TOKEN_CHI; }
	YY_BREAK
case 45:
YY_RULE_SETUP
#line 260 "giza-scanner.l"
{ return GIZA_TOKEN_PSI; }
	YY_BREAK
case 46:
YY_RULE_SETUP
#line 261 "giza-scanner.l"
{ return GIZA_TOKEN_OMEGA; }
	YY_BREAK
case 47:
YY_RULE_SETUP
#line 263 "giza-scanner.l"
{ return GIZA_TOKEN_BIGALPHA; }
	YY_BREAK
case 48:
YY_RULE_SETUP
#line 264 "giza-scanner.l"
{ return GIZA_TOKEN_BIGBETA; }
	YY_BREAK
case 49:
YY_RULE_SETUP
#line 265 "giza-scanner.l"
{ return GIZA_TOKEN_BIGGAMMA; }
	YY_BREAK
case 50:
YY_RULE_SETUP
#line 266 "giza-scanner.l"
{ return GIZA_TOKEN_BIGDELTA; }
	YY_BREAK
case 51:
YY_RULE_SETUP
#line 267 "giza-scanner.l"
{ return GIZA_TOKEN_BIGEPSILON; }
	YY_BREAK
case 52:
YY_RULE_SETUP
#line 268 "giza-scanner.l"
{ return GIZA_TOKEN_BIGZETA; }
	YY_BREAK
case 53:
YY_RULE_SETUP
#line 269 "giza-scanner.l"
{ return GIZA_TOKEN_BIGETA; }
	YY_BREAK
case 54:
YY_RULE_SETUP
#line 270 "giza-scanner.l"
{ return GIZA_TOKEN_BIGTHETA; }
	YY_BREAK
case 55:
YY_RULE_SETUP
#line 271 "giza-scanner.l"
{ return GIZA_TOKEN_BIGIOTA; }
	YY_BREAK
case 56:
YY_RULE_SETUP
#line 272 "giza-scanner.l"
{ return GIZA_TOKEN_BIGKAPPA; }
	YY_BREAK
case 57:
YY_RULE_SETUP
#line 273 "giza-scanner.l"
{ return GIZA_TOKEN_BIGLAMBDA; }
	YY_BREAK
case 58:
YY_RULE_SETUP
#line 274 "giza-scanner.l"
{ return GIZA_TOKEN_BIGMU; }
	YY_BREAK
case 59:
YY_RULE_SETUP
#line 275 "giza-scanner.l"
{ return GIZA_TOKEN_BIGNU; }
	YY_BREAK
case 60:
YY_RULE_SETUP
#line 276 "giza-scanner.l"
{ return GIZA_TOKEN_BIGXI; }
	YY_BREAK
case 61:
YY_RULE_SETUP
#line 277 "giza-scanner.l"
{ return GIZA_TOKEN_BIGOMICRON; }
	YY_BREAK
case 62:
YY_RULE_SETUP
#line 278 "giza-scanner.l"
{ return GIZA_TOKEN_BIGPI; }
	YY_BREAK
case 63:
YY_RULE_SETUP
#line 279 "giza-scanner.l"
{ return GIZA_TOKEN_BIGRHO; }
	YY_BREAK
case 64:
YY_RULE_SETUP
#line 280 "giza-scanner.l"
{ return GIZA_TOKEN_BIGSIGMA; }
	YY_BREAK
case 65:
YY_RULE_SETUP
#line 281 "giza-scanner.l"
{ return GIZA_TOKEN_BIGTAU; }
	YY_BREAK
case 66:
YY_RULE_SETUP
#line 282 "giza-scanner.l"
{ return GIZA_TOKEN_BIGUPSILON; }
	YY_BREAK
case 67:
YY_RULE_SETUP
#line 283 "giza-scanner.l"
{ return GIZA_TOKEN_BIGPHI; }
	YY_BREAK
case 68:
YY_RULE_SETUP
#line 284 "giza-scanner.l"
{ return GIZA_TOKEN_BIGCHI; }
	YY_BREAK
case 69:
YY_RULE_SETUP
#line 285 "giza-scanner.l"
{ return GIZA_TOKEN_BIGPSI; }
	YY_BREAK
case 70:
YY_RULE_SETUP
#line 286 "giza-scanner.l"
{ return GIZA_TOKEN_BIGOMEGA; }
	YY_BREAK
case YY_STATE_EOF(INITIAL):
#line 289 "giza-scanner.l"
{ return GIZA_TOKEN_END; }
	YY_BREAK
case 71:
/* rule 71 can match eol */
YY_RULE_SETUP
#line 291 "giza-scanner.l"
{ return GIZA_TOKEN_ERROR; }
	YY_BREAK
case 72:
YY_RULE_SETUP
#line 293 "giza-scanner.l"
ECHO;
	YY_BREAK
#line 1402 "lex.yy.c"

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - yyg->yytext_ptr) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = yyg->yy_hold_char;
		YY_RESTORE_YY_MORE_OFFSET

		if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed yyin at a new source and called
			 * yylex().  If so, then we have to assure
			 * consistency between YY_CURRENT_BUFFER and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
			YY_CURRENT_BUFFER_LVALUE->yy_input_file = yyin;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( yyg->yy_c_buf_p <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			yyg->yy_c_buf_p = yyg->yytext_ptr + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state( yyscanner );

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state , yyscanner);

			yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++yyg->yy_c_buf_p;
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
				yy_cp = yyg->yy_c_buf_p;
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer( yyscanner ) )
			{
			case EOB_ACT_END_OF_FILE:
				{
				yyg->yy_did_buffer_switch_on_eof = 0;

				if ( yywrap(yyscanner ) )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * yytext, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					yyg->yy_c_buf_p = yyg->yytext_ptr + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				yyg->yy_c_buf_p =
					yyg->yytext_ptr + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				yyg->yy_c_buf_p =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars];

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
} /* end of yylex */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
static int yy_get_next_buffer (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	register char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
	register char *source = yyg->yytext_ptr;
	register int number_to_move, i;
	int ret_val;

	if ( yyg->yy_c_buf_p > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( yyg->yy_c_buf_p - yyg->yytext_ptr - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr) - 1;

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars = 0;

	else
		{
			yy_size_t num_to_read =
			YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = YY_CURRENT_BUFFER_LVALUE;

			int yy_c_buf_p_offset =
				(int) (yyg->yy_c_buf_p - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				yy_size_t new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					yyrealloc((void *) b->yy_ch_buf,b->yy_buf_size + 2 ,yyscanner );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = 0;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			yyg->yy_c_buf_p = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size -
						number_to_move - 1;

			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]),
			yyg->yy_n_chars, num_to_read );

		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	if ( yyg->yy_n_chars == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			yyrestart(yyin  ,yyscanner);
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	if ((yy_size_t) (yyg->yy_n_chars + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size) {
		/* Extend the array by 50%, plus the number we really need. */
		yy_size_t new_size = yyg->yy_n_chars + number_to_move + (yyg->yy_n_chars >> 1);
		YY_CURRENT_BUFFER_LVALUE->yy_ch_buf = (char *) yyrealloc((void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf,new_size ,yyscanner );
		if ( ! YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			YY_FATAL_ERROR( "out of dynamic memory in yy_get_next_buffer()" );
	}

	yyg->yy_n_chars += number_to_move;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] = YY_END_OF_BUFFER_CHAR;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;

	yyg->yytext_ptr = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

	return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

    static yy_state_type yy_get_previous_state (yyscan_t yyscanner)
{
	register yy_state_type yy_current_state;
	register char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yy_current_state = yyg->yy_start;

	for ( yy_cp = yyg->yytext_ptr + YY_MORE_ADJ; yy_cp < yyg->yy_c_buf_p; ++yy_cp )
		{
		register YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 1);
		if ( yy_accept[yy_current_state] )
			{
			yyg->yy_last_accepting_state = yy_current_state;
			yyg->yy_last_accepting_cpos = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 241 )
				yy_c = yy_meta[(unsigned int) yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
		}

	return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
    static yy_state_type yy_try_NUL_trans  (yy_state_type yy_current_state , yyscan_t yyscanner)
{
	register int yy_is_jam;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner; /* This var may be unused depending upon options. */
	register char *yy_cp = yyg->yy_c_buf_p;

	register YY_CHAR yy_c = 1;
	if ( yy_accept[yy_current_state] )
		{
		yyg->yy_last_accepting_state = yy_current_state;
		yyg->yy_last_accepting_cpos = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 241 )
			yy_c = yy_meta[(unsigned int) yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
	yy_is_jam = (yy_current_state == 240);

	(void)yyg;
	return yy_is_jam ? 0 : yy_current_state;
}

    static void yyunput (int c, register char * yy_bp , yyscan_t yyscanner)
{
	register char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    yy_cp = yyg->yy_c_buf_p;

	/* undo effects of setting up yytext */
	*yy_cp = yyg->yy_hold_char;

	if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
		{ /* need to shift things up to make room */
		/* +2 for EOB chars. */
		register yy_size_t number_to_move = yyg->yy_n_chars + 2;
		register char *dest = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[
					YY_CURRENT_BUFFER_LVALUE->yy_buf_size + 2];
		register char *source =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move];

		while ( source > YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			*--dest = *--source;

		yy_cp += (int) (dest - source);
		yy_bp += (int) (dest - source);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars =
			yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_buf_size;

		if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
			YY_FATAL_ERROR( "flex scanner push-back overflow" );
		}

	*--yy_cp = (char) c;

	yyg->yytext_ptr = yy_bp;
	yyg->yy_hold_char = *yy_cp;
	yyg->yy_c_buf_p = yy_cp;
}

#ifndef YY_NO_INPUT
#ifdef __cplusplus
    static int yyinput (yyscan_t yyscanner)
#else
    static int input  (yyscan_t yyscanner)
#endif

{
	int c;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	*yyg->yy_c_buf_p = yyg->yy_hold_char;

	if ( *yyg->yy_c_buf_p == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( yyg->yy_c_buf_p < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			/* This was really a NUL. */
			*yyg->yy_c_buf_p = '\0';

		else
			{ /* need more input */
			yy_size_t offset = yyg->yy_c_buf_p - yyg->yytext_ptr;
			++yyg->yy_c_buf_p;

			switch ( yy_get_next_buffer( yyscanner ) )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					yyrestart(yyin ,yyscanner);

					/*FALLTHROUGH*/

				case EOB_ACT_END_OF_FILE:
					{
					if ( yywrap(yyscanner ) )
						return EOF;

					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput(yyscanner);
#else
					return input(yyscanner);
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					yyg->yy_c_buf_p = yyg->yytext_ptr + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) yyg->yy_c_buf_p;	/* cast for 8-bit char's */
	*yyg->yy_c_buf_p = '\0';	/* preserve yytext */
	yyg->yy_hold_char = *++yyg->yy_c_buf_p;

	return c;
}
#endif	/* ifndef YY_NO_INPUT */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * @param yyscanner The scanner object.
 * @note This function does not reset the start condition to @c INITIAL .
 */
    void yyrestart  (FILE * input_file , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! YY_CURRENT_BUFFER ){
        yyensure_buffer_stack (yyscanner);
		YY_CURRENT_BUFFER_LVALUE =
            yy_create_buffer(yyin,YY_BUF_SIZE ,yyscanner);
	}

	yy_init_buffer(YY_CURRENT_BUFFER,input_file ,yyscanner);
	yy_load_buffer_state(yyscanner );
}

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * @param yyscanner The scanner object.
 */
    void yy_switch_to_buffer  (YY_BUFFER_STATE  new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	/* TODO. We should be able to replace this entire function body
	 * with
	 *		yypop_buffer_state();
	 *		yypush_buffer_state(new_buffer);
     */
	yyensure_buffer_stack (yyscanner);
	if ( YY_CURRENT_BUFFER == new_buffer )
		return;

	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	YY_CURRENT_BUFFER_LVALUE = new_buffer;
	yy_load_buffer_state(yyscanner );

	/* We don't actually know whether we did this switch during
	 * EOF (yywrap()) processing, but the only time this flag
	 * is looked at is after yywrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	yyg->yy_did_buffer_switch_on_eof = 1;
}

static void yy_load_buffer_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
	yyg->yytext_ptr = yyg->yy_c_buf_p = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
	yyin = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
	yyg->yy_hold_char = *yyg->yy_c_buf_p;
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * @param yyscanner The scanner object.
 * @return the allocated buffer state.
 */
    YY_BUFFER_STATE yy_create_buffer  (FILE * file, int  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	b = (YY_BUFFER_STATE) yyalloc(sizeof( struct yy_buffer_state ) ,yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) yyalloc(b->yy_buf_size + 2 ,yyscanner );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_is_our_buffer = 1;

	yy_init_buffer(b,file ,yyscanner);

	return b;
}

/** Destroy the buffer.
 * @param b a buffer created with yy_create_buffer()
 * @param yyscanner The scanner object.
 */
    void yy_delete_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! b )
		return;

	if ( b == YY_CURRENT_BUFFER ) /* Not sure if we should pop here. */
		YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) 0;

	if ( b->yy_is_our_buffer )
		yyfree((void *) b->yy_ch_buf ,yyscanner );

	yyfree((void *) b ,yyscanner );
}

/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a yyrestart() or at EOF.
 */
    static void yy_init_buffer  (YY_BUFFER_STATE  b, FILE * file , yyscan_t yyscanner)

{
	int oerrno = errno;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yy_flush_buffer(b ,yyscanner);

	b->yy_input_file = file;
	b->yy_fill_buffer = 1;

    /* If b is the current buffer, then yy_init_buffer was _probably_
     * called from yyrestart() or through yy_get_next_buffer.
     * In that case, we don't want to reset the lineno or column.
     */
    if (b != YY_CURRENT_BUFFER){
        b->yy_bs_lineno = 1;
        b->yy_bs_column = 0;
    }

        b->yy_is_interactive = file ? (isatty( fileno(file) ) > 0) : 0;
    
	errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * @param yyscanner The scanner object.
 */
    void yy_flush_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == YY_CURRENT_BUFFER )
		yy_load_buffer_state(yyscanner );
}

/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  @param yyscanner The scanner object.
 */
void yypush_buffer_state (YY_BUFFER_STATE new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (new_buffer == NULL)
		return;

	yyensure_buffer_stack(yyscanner);

	/* This block is copied from yy_switch_to_buffer. */
	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	/* Only push if top exists. Otherwise, replace top. */
	if (YY_CURRENT_BUFFER)
		yyg->yy_buffer_stack_top++;
	YY_CURRENT_BUFFER_LVALUE = new_buffer;

	/* copied from yy_switch_to_buffer. */
	yy_load_buffer_state(yyscanner );
	yyg->yy_did_buffer_switch_on_eof = 1;
}

/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  @param yyscanner The scanner object.
 */
void yypop_buffer_state (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (!YY_CURRENT_BUFFER)
		return;

	yy_delete_buffer(YY_CURRENT_BUFFER ,yyscanner);
	YY_CURRENT_BUFFER_LVALUE = NULL;
	if (yyg->yy_buffer_stack_top > 0)
		--yyg->yy_buffer_stack_top;

	if (YY_CURRENT_BUFFER) {
		yy_load_buffer_state(yyscanner );
		yyg->yy_did_buffer_switch_on_eof = 1;
	}
}

/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
static void yyensure_buffer_stack (yyscan_t yyscanner)
{
	yy_size_t num_to_alloc;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if (!yyg->yy_buffer_stack) {

		/* First allocation is just for 2 elements, since we don't know if this
		 * scanner will even need a stack. We use 2 instead of 1 to avoid an
		 * immediate realloc on the next call.
         */
		num_to_alloc = 1;
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yyalloc
								(num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );
								  
		memset(yyg->yy_buffer_stack, 0, num_to_alloc * sizeof(struct yy_buffer_state*));
				
		yyg->yy_buffer_stack_max = num_to_alloc;
		yyg->yy_buffer_stack_top = 0;
		return;
	}

	if (yyg->yy_buffer_stack_top >= (yyg->yy_buffer_stack_max) - 1){

		/* Increase the buffer to prepare for a possible push. */
		int grow_size = 8 /* arbitrary grow size */;

		num_to_alloc = yyg->yy_buffer_stack_max + grow_size;
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yyrealloc
								(yyg->yy_buffer_stack,
								num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );

		/* zero only the new slots.*/
		memset(yyg->yy_buffer_stack + yyg->yy_buffer_stack_max, 0, grow_size * sizeof(struct yy_buffer_state*));
		yyg->yy_buffer_stack_max = num_to_alloc;
	}
}

/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object. 
 */
YY_BUFFER_STATE yy_scan_buffer  (char * base, yy_size_t  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	if ( size < 2 ||
	     base[size-2] != YY_END_OF_BUFFER_CHAR ||
	     base[size-1] != YY_END_OF_BUFFER_CHAR )
		/* They forgot to leave room for the EOB's. */
		return 0;

	b = (YY_BUFFER_STATE) yyalloc(sizeof( struct yy_buffer_state ) ,yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_buffer()" );

	b->yy_buf_size = size - 2;	/* "- 2" to take care of EOB's */
	b->yy_buf_pos = b->yy_ch_buf = base;
	b->yy_is_our_buffer = 0;
	b->yy_input_file = 0;
	b->yy_n_chars = b->yy_buf_size;
	b->yy_is_interactive = 0;
	b->yy_at_bol = 1;
	b->yy_fill_buffer = 0;
	b->yy_buffer_status = YY_BUFFER_NEW;

	yy_switch_to_buffer(b ,yyscanner );

	return b;
}

/** Setup the input buffer state to scan a string. The next call to yylex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       yy_scan_bytes() instead.
 */
YY_BUFFER_STATE yy_scan_string (yyconst char * yystr , yyscan_t yyscanner)
{
    
	return yy_scan_bytes(yystr,strlen(yystr) ,yyscanner);
}

/** Setup the input buffer state to scan the given bytes. The next call to yylex() will
 * scan from a @e copy of @a bytes.
 * @param yybytes the byte buffer to scan
 * @param _yybytes_len the number of bytes in the buffer pointed to by @a bytes.
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yy_scan_bytes  (yyconst char * yybytes, yy_size_t  _yybytes_len , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
	char *buf;
	yy_size_t n;
	int i;
    
	/* Get memory for full buffer, including space for trailing EOB's. */
	n = _yybytes_len + 2;
	buf = (char *) yyalloc(n ,yyscanner );
	if ( ! buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_bytes()" );

	for ( i = 0; i < _yybytes_len; ++i )
		buf[i] = yybytes[i];

	buf[_yybytes_len] = buf[_yybytes_len+1] = YY_END_OF_BUFFER_CHAR;

	b = yy_scan_buffer(buf,n ,yyscanner);
	if ( ! b )
		YY_FATAL_ERROR( "bad buffer in yy_scan_bytes()" );

	/* It's okay to grow etc. this buffer, and we should throw it
	 * away when we're done.
	 */
	b->yy_is_our_buffer = 1;

	return b;
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

static void yy_fatal_error (yyconst char* msg , yyscan_t yyscanner)
{
    	(void) fprintf( stderr, "%s\n", msg );
	exit( YY_EXIT_FAILURE );
}

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		yytext[yyleng] = yyg->yy_hold_char; \
		yyg->yy_c_buf_p = yytext + yyless_macro_arg; \
		yyg->yy_hold_char = *yyg->yy_c_buf_p; \
		*yyg->yy_c_buf_p = '\0'; \
		yyleng = yyless_macro_arg; \
		} \
	while ( 0 )

/* Accessor  methods (get/set functions) to struct members. */

/** Get the user-defined data for this scanner.
 * @param yyscanner The scanner object.
 */
YY_EXTRA_TYPE yyget_extra  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyextra;
}

/** Get the current line number.
 * @param yyscanner The scanner object.
 */
int yyget_lineno  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    
        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yylineno;
}

/** Get the current column number.
 * @param yyscanner The scanner object.
 */
int yyget_column  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    
        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yycolumn;
}

/** Get the input stream.
 * @param yyscanner The scanner object.
 */
FILE *yyget_in  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyin;
}

/** Get the output stream.
 * @param yyscanner The scanner object.
 */
FILE *yyget_out  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyout;
}

/** Get the length of the current token.
 * @param yyscanner The scanner object.
 */
yy_size_t yyget_leng  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyleng;
}

/** Get the current token.
 * @param yyscanner The scanner object.
 */

char *yyget_text  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yytext;
}

/** Set the user-defined data. This data is never touched by the scanner.
 * @param user_defined The data to be associated with this scanner.
 * @param yyscanner The scanner object.
 */
void yyset_extra (YY_EXTRA_TYPE  user_defined , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyextra = user_defined ;
}

/** Set the current line number.
 * @param line_number
 * @param yyscanner The scanner object.
 */
void yyset_lineno (int  line_number , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* lineno is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           YY_FATAL_ERROR( "yyset_lineno called with no buffer" );
    
    yylineno = line_number;
}

/** Set the current column.
 * @param line_number
 * @param yyscanner The scanner object.
 */
void yyset_column (int  column_no , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* column is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           YY_FATAL_ERROR( "yyset_column called with no buffer" );
    
    yycolumn = column_no;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param in_str A readable stream.
 * @param yyscanner The scanner object.
 * @see yy_switch_to_buffer
 */
void yyset_in (FILE *  in_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyin = in_str ;
}

void yyset_out (FILE *  out_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyout = out_str ;
}

int yyget_debug  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yy_flex_debug;
}

void yyset_debug (int  bdebug , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yy_flex_debug = bdebug ;
}

/* Accessor methods for yylval and yylloc */

/* User-visible API */

/* yylex_init is special because it creates the scanner itself, so it is
 * the ONLY reentrant function that doesn't take the scanner as the last argument.
 * That's why we explicitly handle the declaration, instead of using our macros.
 */

int yylex_init(yyscan_t* ptr_yy_globals)

{
    if (ptr_yy_globals == NULL){
        errno = EINVAL;
        return 1;
    }

    *ptr_yy_globals = (yyscan_t) yyalloc ( sizeof( struct yyguts_t ), NULL );

    if (*ptr_yy_globals == NULL){
        errno = ENOMEM;
        return 1;
    }

    /* By setting to 0xAA, we expose bugs in yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));

    return yy_init_globals ( *ptr_yy_globals );
}

/* yylex_init_extra has the same functionality as yylex_init, but follows the
 * convention of taking the scanner as the last argument. Note however, that
 * this is a *pointer* to a scanner, as it will be allocated by this call (and
 * is the reason, too, why this function also must handle its own declaration).
 * The user defined value in the first argument will be available to yyalloc in
 * the yyextra field.
 */

int yylex_init_extra(YY_EXTRA_TYPE yy_user_defined,yyscan_t* ptr_yy_globals )

{
    struct yyguts_t dummy_yyguts;

    yyset_extra (yy_user_defined, &dummy_yyguts);

    if (ptr_yy_globals == NULL){
        errno = EINVAL;
        return 1;
    }
	
    *ptr_yy_globals = (yyscan_t) yyalloc ( sizeof( struct yyguts_t ), &dummy_yyguts );
	
    if (*ptr_yy_globals == NULL){
        errno = ENOMEM;
        return 1;
    }
    
    /* By setting to 0xAA, we expose bugs in
    yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));
    
    yyset_extra (yy_user_defined, *ptr_yy_globals);
    
    return yy_init_globals ( *ptr_yy_globals );
}

static int yy_init_globals (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    /* Initialization is the same as for the non-reentrant scanner.
     * This function is called from yylex_destroy(), so don't allocate here.
     */

    yyg->yy_buffer_stack = 0;
    yyg->yy_buffer_stack_top = 0;
    yyg->yy_buffer_stack_max = 0;
    yyg->yy_c_buf_p = (char *) 0;
    yyg->yy_init = 0;
    yyg->yy_start = 0;

    yyg->yy_start_stack_ptr = 0;
    yyg->yy_start_stack_depth = 0;
    yyg->yy_start_stack =  NULL;

/* Defined in main.c */
#ifdef YY_STDINIT
    yyin = stdin;
    yyout = stdout;
#else
    yyin = (FILE *) 0;
    yyout = (FILE *) 0;
#endif

    /* For future reference: Set errno on error, since we are called by
     * yylex_init()
     */
    return 0;
}

/* yylex_destroy is for both reentrant and non-reentrant scanners. */
int yylex_destroy  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    /* Pop the buffer stack, destroying each element. */
	while(YY_CURRENT_BUFFER){
		yy_delete_buffer(YY_CURRENT_BUFFER ,yyscanner );
		YY_CURRENT_BUFFER_LVALUE = NULL;
		yypop_buffer_state(yyscanner);
	}

	/* Destroy the stack itself. */
	yyfree(yyg->yy_buffer_stack ,yyscanner);
	yyg->yy_buffer_stack = NULL;

    /* Destroy the start condition stack. */
        yyfree(yyg->yy_start_stack ,yyscanner );
        yyg->yy_start_stack = NULL;

    /* Reset the globals. This is important in a non-reentrant scanner so the next time
     * yylex() is called, initialization will occur. */
    yy_init_globals( yyscanner);

    /* Destroy the main struct (reentrant only). */
    yyfree ( yyscanner , yyscanner );
    yyscanner = NULL;
    return 0;
}

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void yy_flex_strncpy (char* s1, yyconst char * s2, int n , yyscan_t yyscanner)
{
	register int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * s , yyscan_t yyscanner)
{
	register int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
}
#endif

void *yyalloc (yy_size_t  size , yyscan_t yyscanner)
{
	return (void *) malloc( size );
}

void *yyrealloc  (void * ptr, yy_size_t  size , yyscan_t yyscanner)
{
	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
	return (void *) realloc( (char *) ptr, size );
}

void yyfree (void * ptr , yyscan_t yyscanner)
{
	free( (char *) ptr );	/* see yyrealloc() for (char *) cast */
}

#define YYTABLES_NAME "yytables"

#line 293 "giza-scanner.l"



/**
 * Parses brackets in text, returns the length of the string contained in outermost brackets
 */
int
_giza_get_chunk (const char *text)
{
  int openb = 0;
  int length = 0;
  int token, stop = 0;

  /* initialise the scanner */
  yyscan_t scanner;
  yylex_init (&scanner);
  yy_scan_string (text, scanner);
 
  while(!stop)
    {
      token = yylex(scanner);
      length += yyget_leng (scanner);

      switch (token)
        {
	  case GIZA_TOKEN_OB:
	    ++openb;
	    break;
	  case GIZA_TOKEN_CB:
	    --openb;
	    break;
          case GIZA_TOKEN_END:
            stop = 1;
            break;
	  default:
	    break;
        }

	if (openb < 1) 
          {
            yylex_destroy (scanner);
            return length;
          }
    }
   yylex_destroy (scanner);
   return length; /* assumes whole string if unclosed open bracket */
}

/**
 * Parses remaining text to determine the chunk of text to which the font 
 * change should apply. In this case the chunk is defined as the text 
 * remaining until an unmatched close bracket, or the end of the string, 
 * whichever comes first.
 */
int
_giza_get_chunk_fontchange (const char *text)
{
  int openb = 0;
  int length = 0;
  int token, stop = 0;

  /* initialise the scanner */
  yyscan_t scanner;
  yylex_init (&scanner);
  yy_scan_string (text, scanner);
 
  while(!stop)
    {
      token = yylex(scanner);
      length += yyget_leng (scanner);

      switch (token)
        {
	  case GIZA_TOKEN_OB:
	    ++openb;
	    break;
	  case GIZA_TOKEN_CB:
	    --openb;
	    break;
          case GIZA_TOKEN_END:
            stop = 1;
            break;
	  default:
	    break;
        }

        if (openb < 0) {
           stop = 1;
        }
    }
   yylex_destroy (scanner);
   if (strlen(text) < length) length = strlen(text);
   return length; /* assumes whole string if unclosed open bracket */
}

/**
 * Performs action on each chunk of text. If calculating size height and width get set to height and width of the text.
 */
void
_giza_parse_string (const char *text, double *width, double *height, void (*action)(const char *, double *, double *))
{
  yyscan_t scanner;
  int stop = 0, token, chunkLength, oldTrans;
  int posInText = 0;
  int insuper = 0;
  int insub   = 0;
  int lenstr  = strlen(text);
  char processed[lenstr + 1];
  processed[0] = '\0';

  oldTrans = _giza_get_trans ();
  _giza_set_trans (GIZA_TRANS_IDEN);
  int lenfont = GIZA_FONT_LEN;
  char oldfont[lenfont];

  /* initialise the scanner */
  yylex_init (&scanner);
  yy_scan_string (text, scanner);
 
  while(!stop)
    {
      token = yylex(scanner);
      posInText += yyget_leng (scanner);;
      /*printf("in parse string, token = %i pos = %i processed=\"%s\"\n",token,posInText,processed); */

      switch(token)
        {
        case GIZA_TOKEN_OB:
        case GIZA_TOKEN_CB: /* ignore closing bracket! */
          break;
        case GIZA_TOKEN_SUPER:
          /* Perform action on the text so far parsed */
          action (processed, width, height);
          processed[0] = '\0';

          /* if end of string, do nothing */
          if (posInText >= lenstr)
            { 
              stop = 1;
              break;
            }

          /* Parse the brackets, i.e. get the chunk of text that needs to be superscript */
          chunkLength = _giza_get_chunk (text + posInText);
          strncpy (processed, text + posInText, chunkLength);
          processed[chunkLength] = '\0';
          
          /* Perform the action on this chunk */
          _giza_start_super ();
          _giza_parse_string (processed, width, height, action);
          _giza_stop_super ();

          /* restart the scanner at the new position! */
          yylex_destroy (scanner);
          yylex_init (&scanner);
          yy_scan_string (text + posInText + chunkLength, scanner);

          posInText += chunkLength;

          processed[0] = '\0';
          break;
        case GIZA_TOKEN_SUB:
          /* Perform action on text so far parsed */
          action (processed, width, height);
          processed[0] = '\0';

          /* if end of string, do nothing */
          if (posInText >= lenstr)
            { 
              stop = 1;
              break;
            }

          /* Parse the brackets */
          chunkLength = _giza_get_chunk (text + posInText);
          strncpy (processed, text + posInText, chunkLength);
          processed[chunkLength] = '\0';

          /* perform action on this chunk */
          _giza_start_sub ();
          _giza_parse_string (processed, width, height, action);
          _giza_stop_sub ();

          /* restart the scanner at the new position! */
          yylex_destroy (scanner);
          yylex_init (&scanner);
          yy_scan_string (text + posInText + chunkLength, scanner);

          posInText += chunkLength;

          processed[0] = '\0';
          break;
        case GIZA_TOKEN_RAISE:
          /* Perform action on the text so far parsed */
          action (processed, width, height);
          processed[0] = '\0';
          if (insub > 0)
            {
             _giza_stop_sub();
             insub -= 1;
            }
          else
            {
             _giza_start_super();
             insuper += 1;
            }
          break;
        case GIZA_TOKEN_LOWER:
          /* Perform action on the text so far parsed */
          action (processed, width, height);
          processed[0] = '\0';
          if (insuper > 0)
            {
             _giza_stop_super();
             insuper -= 1;
            }
          else
            {
             _giza_start_sub();
             insub += 1;
            }
          break;
        case GIZA_TOKEN_FONT_NORMAL:
        case GIZA_TOKEN_FONT_ROMAN:
        case GIZA_TOKEN_FONT_ITALIC:
        case GIZA_TOKEN_FONT_SCRIPT:
        case GIZA_TOKEN_FONT_BOLD:
          /* Perform action on the text so far parsed */
          action (processed, width, height);
          processed[0] = '\0';

          /* if end of string, do nothing */
          if (posInText >= lenstr)
            { 
              stop = 1;
              break;
            }

          /* Parse the brackets */
          chunkLength = _giza_get_chunk_fontchange (text + posInText);
          strncpy (processed, text + posInText, chunkLength);
          processed[chunkLength] = '\0';
          
          /* save font */
          giza_get_font(oldfont,lenfont);

          /* perform action on this chunk */
          switch (token)
            {
            case GIZA_TOKEN_FONT_BOLD:
              _giza_switch_font (5);
              break;
            case GIZA_TOKEN_FONT_SCRIPT:
              _giza_switch_font (4);
              break;
            case GIZA_TOKEN_FONT_ITALIC:
              _giza_switch_font (3);
              break;
            case GIZA_TOKEN_FONT_ROMAN:
              _giza_switch_font (2);
              break;
            case GIZA_TOKEN_FONT_NORMAL:
              _giza_switch_font (1);
              break;
            }
          _giza_parse_string (processed, width, height, action);

          /* restart the scanner at the new position! */
          yylex_destroy (scanner);
          yylex_init (&scanner);
          yy_scan_string (text + posInText + chunkLength, scanner);

          posInText += chunkLength;

          /* restore font */
          giza_set_font(oldfont);

          processed[0] = '\0';
          break;
        case GIZA_TOKEN_MARKER:
          /* Perform action on the text so far parsed */
          action (processed, width, height);

          /* if end of string, do nothing */
          if (posInText >= lenstr)
            { 
              stop = 1;
              break;
            }

          /* get current marker height */
          double markerHeight;
          _giza_get_markerheight(&markerHeight);

          /* extract the symbol number from in front of the \m */
          int number;
          sscanf(text+posInText,"%i",&number);
          
          /* 
           * get the length of the string containing the number 
           * so we know how many characters we have processed
           */
          char temp[4];
          sprintf(temp,"%i",number);
          chunkLength = strlen(temp);
          
          /* either draw the symbol or increment the width */
          if (*width < 0.) {
             double xp, yp;
             double sina = sin(Dev[id].fontAngle);
             double cosa = cos(Dev[id].fontAngle);
             double dx = 0.75*markerHeight;
             double dy = -0.5*markerHeight;
             cairo_get_current_point(Dev[id].context, &xp, &yp); /* current pen position */
             _giza_draw_symbol_device(xp + dx*cosa + dy*sina, yp + dx*sina + dy*cosa, number);
             cairo_move_to(Dev[id].context, xp + 2.*dx*cosa, yp + 2.*dx*sina); /* restore pen position */

          } else {
             /* return width if action is _giza_action_get_size */
             *width = *width + 1.5*markerHeight;
          }
          

          /* restart the scanner at the new position! */
          yylex_destroy (scanner);
          yylex_init (&scanner);
          yy_scan_string (text + posInText + chunkLength, scanner);

          posInText += chunkLength;

          processed[0] = '\0';
          break;

        case GIZA_TOKEN_END:
          stop = 1;
          break;
        default:
          _giza_add_token_to_string (processed, scanner, token);
          break;
        }
    }
  yylex_destroy (scanner);

  /* Perform action on the last section of text */
  action (processed, width, height);

  _giza_set_trans (oldTrans);
}

/**
 * Adds the utf-8 character representation of token to string.
 */
void
_giza_add_token_to_string (char *string, yyscan_t scanner, int token)
{
  switch (token)
    {
     case GIZA_TOKEN_OTHER: /* print the ascii or unicode char */
       strcat (string, yyget_text (scanner));
       break;
     case GIZA_TOKEN_TIMES:
       strcat (string, "×");
       break;
     case GIZA_TOKEN_ALPHA: 
       strcat (string, "α");
       break;
     case GIZA_TOKEN_BETA: 
       strcat (string, "β");
       break;
     case GIZA_TOKEN_GAMMA: 
       strcat (string, "γ");
       break;
     case GIZA_TOKEN_DELTA: 
       strcat (string, "δ");
       break;
     case GIZA_TOKEN_EPSILON: 
       strcat (string, "ε");
       break;
     case GIZA_TOKEN_ZETA: 
       strcat (string, "ζ");
       break;
     case GIZA_TOKEN_ETA: 
       strcat (string, "η");
       break;
     case GIZA_TOKEN_THETA: 
       strcat (string, "θ");
       break;
     case GIZA_TOKEN_IOTA: 
       strcat (string, "ι");
       break;
     case GIZA_TOKEN_KAPPA: 
       strcat (string, "κ");
       break;
     case GIZA_TOKEN_LAMBDA: 
       strcat (string, "λ");
       break;
     case GIZA_TOKEN_MU: 
       strcat (string, "μ");
       break;
     case GIZA_TOKEN_NU: 
       strcat (string, "ν");
       break;
     case GIZA_TOKEN_XI: 
       strcat (string, "ξ");
       break;
     case GIZA_TOKEN_OMICRON: 
       strcat (string, "ο");
       break;
     case GIZA_TOKEN_PI: 
       strcat (string, "π");
       break;
     case GIZA_TOKEN_RHO: 
       strcat (string, "ρ");
       break;
     case GIZA_TOKEN_SIGMA: 
       strcat (string, "σ");
       break;
     case GIZA_TOKEN_TAU: 
       strcat (string, "τ");
       break;
     case GIZA_TOKEN_UPSILON: 
       strcat (string, "υ");
       break;
     case GIZA_TOKEN_PHI: 
       strcat (string, "φ");
       break;
     case GIZA_TOKEN_CHI: 
       strcat (string, "χ");
       break;
     case GIZA_TOKEN_PSI: 
       strcat (string, "ψ");
       break;
     case GIZA_TOKEN_OMEGA: 
       strcat (string, "ω");
       break;
     case GIZA_TOKEN_BIGALPHA: 
       strcat (string, "Α");
       break;
     case GIZA_TOKEN_BIGBETA: 
       strcat (string, "Β");
       break;
     case GIZA_TOKEN_BIGGAMMA: 
       strcat (string, "Γ");
       break;
     case GIZA_TOKEN_BIGDELTA: 
       strcat (string, "Δ");
       break;
     case GIZA_TOKEN_BIGEPSILON: 
       strcat (string, "Ε");
       break;
     case GIZA_TOKEN_BIGZETA: 
       strcat (string, "Ζ");
       break;
     case GIZA_TOKEN_BIGETA: 
       strcat (string, "Η");
       break;
     case GIZA_TOKEN_BIGTHETA: 
       strcat (string, "Θ");
       break;
     case GIZA_TOKEN_BIGIOTA: 
       strcat (string, "Ι");
       break;
     case GIZA_TOKEN_BIGKAPPA: 
       strcat (string, "Κ");
       break;
     case GIZA_TOKEN_BIGLAMBDA: 
       strcat (string, "Λ");
       break;
     case GIZA_TOKEN_BIGMU: 
       strcat (string, "Μ");
       break;
     case GIZA_TOKEN_BIGNU: 
       strcat (string, "Ν");
       break;
     case GIZA_TOKEN_BIGXI: 
       strcat (string, "Ξ");
       break;
     case GIZA_TOKEN_BIGOMICRON: 
       strcat (string, "Ο");
       break;
     case GIZA_TOKEN_BIGPI: 
       strcat (string, "Π");
       break;
     case GIZA_TOKEN_BIGRHO: 
       strcat (string, "Ρ");
       break;
     case GIZA_TOKEN_BIGSIGMA: 
       strcat (string, "Σ");
       break;
     case GIZA_TOKEN_BIGTAU: 
       strcat (string, "Τ");
       break;
     case GIZA_TOKEN_BIGUPSILON: 
       strcat (string, "Υ");
       break;
     case GIZA_TOKEN_BIGPHI: 
       strcat (string, "Φ");
       break;
     case GIZA_TOKEN_BIGCHI: 
       strcat (string, "Χ");
       break;
     case GIZA_TOKEN_BIGPSI: 
       strcat (string, "Ψ");
       break;
     case GIZA_TOKEN_BIGOMEGA: 
       strcat (string, "Ω");
       break;
     case GIZA_TOKEN_INTEGRAL:
       strcat (string, "∫");
       break;
     case GIZA_TOKEN_NABLA:
       strcat (string, "∇");
       break;
     case GIZA_TOKEN_ODOT:
       strcat (string, "⊙");
       break;
     case GIZA_TOKEN_BACKSLASH:
       strcat (string, "\\");
       break;
     case GIZA_TOKEN_UNDERSCORE:
       strcat (string, "_");
       break;
     case GIZA_TOKEN_HAT:
       strcat (string, "^");
       break;
     case GIZA_TOKEN_CURLYOPEN:
       strcat (string, "{");
       break;
     case GIZA_TOKEN_CURLYCLOSE:
       strcat (string, "}");
       break;
     default:
       break;
    }
}

